"""
    domonic.x3d
    ====================================
    Generate x3d with python 3

"""

from domonic.dom import Element
from domonic.html import closed_tag

x3d = X3D = type("x3d", (Element,), {"name": "x3d"})
scene = Scene = type("scene", (Element,), {"name": "scene"})
material = Material = type("material", (Element,), {"name": "material"})
appearance = Appearance = type("appearance", (Element,), {"name": "appearance"})
sphere = Sphere = type("sphere", (Element,), {"name": "sphere"})
shape = Shape = type("shape", (Element,), {"name": "shape"})
transform = Transform = type("transform", (Element,), {"name": "transform"})
timeSensor = TimeSensor = type("timeSensor", (Element,), {"name": "timeSensor"})
inline = Inline = type("inline", (Element,), {"name": "inline"})
box = Box = type("box", (Element,), {"name": "box"})
plane = Plane = type("plane", (Element,), {"name": "plane"})

# TODO - go through examples to find which are usually lower. i.e before they allowed mised cases
positionInterpolator = PositionInterpolator = type(
    "PositionInterpolator", (Element,), {"name": "PositionInterpolator"}
)
route = Route = type("Route", (Element,), {"name": "Route"})
anchor = Anchor = type("Anchor", (Element,), {"name": "Anchor"})
arc2D = Arc2D = type("Arc2D", (Element,), {"name": "Arc2D"})
arcClose2D = ArcClose2D = type("ArcClose2D", (Element,), {"name": "ArcClose2D"})
audioClip = AudioClip = type("AudioClip", (Element,), {"name": "AudioClip"})
background = Background = type("Background", (Element,), {"name": "Background"})
ballJoint = BallJoint = type("BallJoint", (Element,), {"name": "BallJoint"})
billboard = Billboard = type("Billboard", (Element,), {"name": "Billboard"})
binaryGeometry = BinaryGeometry = type(
    "BinaryGeometry", (Element,), {"name": "BinaryGeometry"}
)
blendedVolumeStyle = BlendedVolumeStyle = type(
    "BlendedVolumeStyle", (Element,), {"name": "BlendedVolumeStyle"}
)
blendMode = BlendMode = type("BlendMode", (Element,), {"name": "BlendMode"})
block = Block = type("Block", (Element,), {"name": "Block"})
boundaryEnhancementVolumeStyle = BoundaryEnhancementVolumeStyle = type(
    "BoundaryEnhancementVolumeStyle",
    (Element,),
    {"name": "BoundaryEnhancementVolumeStyle"},
)
bufferAccessor = BufferAccessor = type(
    "BufferAccessor", (Element,), {"name": "BufferAccessor"}
)
bufferGeometry = BufferGeometry = type(
    "BufferGeometry", (Element,), {"name": "BufferGeometry"}
)
bufferView = BufferView = type("BufferView", (Element,), {"name": "BufferView"})
cADAssembly = CADAssembly = type("CADAssembly", (Element,), {"name": "CADAssembly"})
cADFace = CADFace = type("CADFace", (Element,), {"name": "CADFace"})
cADLayer = CADLayer = type("CADLayer", (Element,), {"name": "CADLayer"})
cADPart = CADPart = type("CADPart", (Element,), {"name": "CADPart"})
cartoonVolumeStyle = CartoonVolumeStyle = type(
    "CartoonVolumeStyle", (Element,), {"name": "CartoonVolumeStyle"}
)
circle2D = Circle2D = type("Circle2D", (Element,), {"name": "Circle2D"})
clipPlane = ClipPlane = type("ClipPlane", (Element,), {"name": "ClipPlane"})
collidableShape = CollidableShape = type(
    "CollidableShape", (Element,), {"name": "CollidableShape"}
)
collision = Collision = type("Collision", (Element,), {"name": "Collision"})
collisionCollection = CollisionCollection = type(
    "CollisionCollection", (Element,), {"name": "CollisionCollection"}
)
collisionSensor = CollisionSensor = type(
    "CollisionSensor", (Element,), {"name": "CollisionSensor"}
)
color = Color = type("Color", (Element,), {"name": "Color"})
colorChaser = ColorChaser = type("ColorChaser", (Element,), {"name": "ColorChaser"})
colorDamper = ColorDamper = type("ColorDamper", (Element,), {"name": "ColorDamper"})
colorInterpolator = ColorInterpolator = type(
    "ColorInterpolator", (Element,), {"name": "ColorInterpolator"}
)
colorMaskMode = ColorMaskMode = type(
    "ColorMaskMode", (Element,), {"name": "ColorMaskMode"}
)
colorRGBA = ColorRGBA = type("ColorRGBA", (Element,), {"name": "ColorRGBA"})
commonSurfaceShader = CommonSurfaceShader = type(
    "CommonSurfaceShader", (Element,), {"name": "CommonSurfaceShader"}
)
composedCubeMapTexture = ComposedCubeMapTexture = type(
    "ComposedCubeMapTexture", (Element,), {"name": "ComposedCubeMapTexture"}
)
composedShader = ComposedShader = type(
    "ComposedShader", (Element,), {"name": "ComposedShader"}
)
composedTexture3D = ComposedTexture3D = type(
    "ComposedTexture3D", (Element,), {"name": "ComposedTexture3D"}
)
composedVolumeStyle = ComposedVolumeStyle = type(
    "ComposedVolumeStyle", (Element,), {"name": "ComposedVolumeStyle"}
)
cone = Cone = type("Cone", (Element,), {"name": "Cone"})
coordinate = Coordinate = type("Coordinate", (Element,), {"name": "Coordinate"})
coordinateDamper = CoordinateDamper = type(
    "CoordinateDamper", (Element,), {"name": "CoordinateDamper"}
)
coordinateDouble = CoordinateDouble = type(
    "CoordinateDouble", (Element,), {"name": "CoordinateDouble"}
)
coordinateInterpolator = CoordinateInterpolator = type(
    "CoordinateInterpolator", (Element,), {"name": "CoordinateInterpolator"}
)
cylinder = Cylinder = type("Cylinder", (Element,), {"name": "Cylinder"})
cylinderSensor = CylinderSensor = type(
    "CylinderSensor", (Element,), {"name": "CylinderSensor"}
)
depthMode = DepthMode = type("DepthMode", (Element,), {"name": "DepthMode"})
directionalLight = DirectionalLight = type(
    "DirectionalLight", (Element,), {"name": "DirectionalLight"}
)
dish = Dish = type("Dish", (Element,), {"name": "Dish"})
disk2D = Disk2D = type("Disk2D", (Element,), {"name": "Disk2D"})
doubleAxisHingeJoint = DoubleAxisHingeJoint = type(
    "DoubleAxisHingeJoint", (Element,), {"name": "DoubleAxisHingeJoint"}
)
dynamicLOD = DynamicLOD = type("DynamicLOD", (Element,), {"name": "DynamicLOD"})
edgeEnhancementVolumeStyle = EdgeEnhancementVolumeStyle = type(
    "EdgeEnhancementVolumeStyle", (Element,), {"name": "EdgeEnhancementVolumeStyle"}
)
elevationGrid = ElevationGrid = type(
    "ElevationGrid", (Element,), {"name": "ElevationGrid"}
)
environment = Environment = type("Environment", (Element,), {"name": "Environment"})
extrusion = Extrusion = type("Extrusion", (Element,), {"name": "Extrusion"})
field = Field = type("Field", (Element,), {"name": "Field"})
floatVertexAttribute = FloatVertexAttribute = type(
    "FloatVertexAttribute", (Element,), {"name": "FloatVertexAttribute"}
)
fog = Fog = type("Fog", (Element,), {"name": "Fog"})
fontStyle = FontStyle = type("FontStyle", (Element,), {"name": "FontStyle"})
generatedCubeMapTexture = GeneratedCubeMapTexture = type(
    "GeneratedCubeMapTexture", (Element,), {"name": "GeneratedCubeMapTexture"}
)
geoCoordinate = GeoCoordinate = type(
    "GeoCoordinate", (Element,), {"name": "GeoCoordinate"}
)
geoElevationGrid = GeoElevationGrid = type(
    "GeoElevationGrid", (Element,), {"name": "GeoElevationGrid"}
)
geoLocation = GeoLocation = type("GeoLocation", (Element,), {"name": "GeoLocation"})
geoLOD = GeoLOD = type("GeoLOD", (Element,), {"name": "GeoLOD"})
geoMetadata = GeoMetadata = type("GeoMetadata", (Element,), {"name": "GeoMetadata"})
geoOrigin = GeoOrigin = type("GeoOrigin", (Element,), {"name": "GeoOrigin"})
geoPositionInterpolator = GeoPositionInterpolator = type(
    "GeoPositionInterpolator", (Element,), {"name": "GeoPositionInterpolator"}
)
geoTransform = GeoTransform = type("GeoTransform", (Element,), {"name": "GeoTransform"})
geoViewpoint = GeoViewpoint = type("GeoViewpoint", (Element,), {"name": "GeoViewpoint"})
group = Group = type("Group", (Element,), {"name": "Group"})
hAnimDisplacer = HAnimDisplacer = type(
    "HAnimDisplacer", (Element,), {"name": "HAnimDisplacer"}
)
hAnimHumanoid = HAnimHumanoid = type(
    "HAnimHumanoid", (Element,), {"name": "HAnimHumanoid"}
)
hAnimJoint = HAnimJoint = type("HAnimJoint", (Element,), {"name": "HAnimJoint"})
hAnimSegment = HAnimSegment = type("HAnimSegment", (Element,), {"name": "HAnimSegment"})
hAnimSite = HAnimSite = type("HAnimSite", (Element,), {"name": "HAnimSite"})
imageTexture = ImageTexture = type("ImageTexture", (Element,), {"name": "ImageTexture"})
imageTexture3D = ImageTexture3D = type(
    "ImageTexture3D", (Element,), {"name": "ImageTexture3D"}
)
imageTextureAtlas = ImageTextureAtlas = type(
    "ImageTextureAtlas", (Element,), {"name": "ImageTextureAtlas"}
)
indexedFaceSet = IndexedFaceSet = type(
    "IndexedFaceSet", (Element,), {"name": "IndexedFaceSet"}
)
indexedLineSet = IndexedLineSet = type(
    "IndexedLineSet", (Element,), {"name": "IndexedLineSet"}
)
indexedQuadSet = IndexedQuadSet = type(
    "IndexedQuadSet", (Element,), {"name": "IndexedQuadSet"}
)
indexedTriangleSet = IndexedTriangleSet = type(
    "IndexedTriangleSet", (Element,), {"name": "IndexedTriangleSet"}
)
indexedTriangleStripSet = IndexedTriangleStripSet = type(
    "IndexedTriangleStripSet", (Element,), {"name": "IndexedTriangleStripSet"}
)
isoSurfaceVolumeData = IsoSurfaceVolumeData = type(
    "IsoSurfaceVolumeData", (Element,), {"name": "IsoSurfaceVolumeData"}
)
lineProperties = LineProperties = type(
    "LineProperties", (Element,), {"name": "LineProperties"}
)
lineSet = LineSet = type("LineSet", (Element,), {"name": "LineSet"})
lOD = LOD = type("LOD", (Element,), {"name": "LOD"})
matrixTextureTransform = MatrixTextureTransform = type(
    "MatrixTextureTransform", (Element,), {"name": "MatrixTextureTransform"}
)
matrixTransform = MatrixTransform = type(
    "MatrixTransform", (Element,), {"name": "MatrixTransform"}
)
mesh = Mesh = type("Mesh", (Element,), {"name": "Mesh"})
metadataBoolean = MetadataBoolean = type(
    "MetadataBoolean", (Element,), {"name": "MetadataBoolean"}
)
metadataDouble = MetadataDouble = type(
    "MetadataDouble", (Element,), {"name": "MetadataDouble"}
)
metadataFloat = MetadataFloat = type(
    "MetadataFloat", (Element,), {"name": "MetadataFloat"}
)
metadataInteger = MetadataInteger = type(
    "MetadataInteger", (Element,), {"name": "MetadataInteger"}
)
metadataSet = MetadataSet = type("MetadataSet", (Element,), {"name": "MetadataSet"})
metadataString = MetadataString = type(
    "MetadataString", (Element,), {"name": "MetadataString"}
)
motorJoint = MotorJoint = type("MotorJoint", (Element,), {"name": "MotorJoint"})
movieTexture = MovieTexture = type("MovieTexture", (Element,), {"name": "MovieTexture"})
mPRPlane = MPRPlane = type("MPRPlane", (Element,), {"name": "MPRPlane"})
mPRVolumeStyle = MPRVolumeStyle = type(
    "MPRVolumeStyle", (Element,), {"name": "MPRVolumeStyle"}
)
multiTexture = MultiTexture = type("MultiTexture", (Element,), {"name": "MultiTexture"})
multiTextureCoordinate = MultiTextureCoordinate = type(
    "MultiTextureCoordinate", (Element,), {"name": "MultiTextureCoordinate"}
)
navigationInfo = NavigationInfo = type(
    "NavigationInfo", (Element,), {"name": "NavigationInfo"}
)
normal = Normal = type("Normal", (Element,), {"name": "Normal"})
normalInterpolator = NormalInterpolator = type(
    "NormalInterpolator", (Element,), {"name": "NormalInterpolator"}
)
nozzle = Nozzle = type("Nozzle", (Element,), {"name": "Nozzle"})
opacityMapVolumeStyle = OpacityMapVolumeStyle = type(
    "OpacityMapVolumeStyle", (Element,), {"name": "OpacityMapVolumeStyle"}
)
orientationChaser = OrientationChaser = type(
    "OrientationChaser", (Element,), {"name": "OrientationChaser"}
)
orientationDamper = OrientationDamper = type(
    "OrientationDamper", (Element,), {"name": "OrientationDamper"}
)
orientationInterpolator = OrientationInterpolator = type(
    "OrientationInterpolator", (Element,), {"name": "OrientationInterpolator"}
)
orthoViewpoint = OrthoViewpoint = type(
    "OrthoViewpoint", (Element,), {"name": "OrthoViewpoint"}
)
param = Param = type("Param", (Element,), {"name": "Param"})
particleSet = ParticleSet = type("ParticleSet", (Element,), {"name": "ParticleSet"})
physicalEnvironmentLight = PhysicalEnvironmentLight = type(
    "PhysicalEnvironmentLight", (Element,), {"name": "PhysicalEnvironmentLight"}
)
physicalMaterial = PhysicalMaterial = type(
    "PhysicalMaterial", (Element,), {"name": "PhysicalMaterial"}
)
pixelTexture = PixelTexture = type("PixelTexture", (Element,), {"name": "PixelTexture"})
pixelTexture3D = PixelTexture3D = type(
    "PixelTexture3D", (Element,), {"name": "PixelTexture3D"}
)
planeSensor = PlaneSensor = type("PlaneSensor", (Element,), {"name": "PlaneSensor"})
pointLight = PointLight = type("PointLight", (Element,), {"name": "PointLight"})
pointSet = PointSet = type("PointSet", (Element,), {"name": "PointSet"})
polyline2D = Polyline2D = type("Polyline2D", (Element,), {"name": "Polyline2D"})
polypoint2D = Polypoint2D = type("Polypoint2D", (Element,), {"name": "Polypoint2D"})
popGeometry = PopGeometry = type("PopGeometry", (Element,), {"name": "PopGeometry"})
popGeometryLevel = PopGeometryLevel = type(
    "PopGeometryLevel", (Element,), {"name": "PopGeometryLevel"}
)
positionChaser = PositionChaser = type(
    "PositionChaser", (Element,), {"name": "PositionChaser"}
)
positionChaser2D = PositionChaser2D = type(
    "PositionChaser2D", (Element,), {"name": "PositionChaser2D"}
)
positionDamper = PositionDamper = type(
    "PositionDamper", (Element,), {"name": "PositionDamper"}
)
positionDamper2D = PositionDamper2D = type(
    "PositionDamper2D", (Element,), {"name": "PositionDamper2D"}
)
positionInterpolator = PositionInterpolator = type(
    "PositionInterpolator", (Element,), {"name": "PositionInterpolator"}
)
positionInterpolator2D = PositionInterpolator2D = type(
    "PositionInterpolator2D", (Element,), {"name": "PositionInterpolator2D"}
)
projectionVolumeStyle = ProjectionVolumeStyle = type(
    "ProjectionVolumeStyle", (Element,), {"name": "ProjectionVolumeStyle"}
)
pyramid = Pyramid = type("Pyramid", (Element,), {"name": "Pyramid"})
quadSet = QuadSet = type("QuadSet", (Element,), {"name": "QuadSet"})
radarVolumeStyle = RadarVolumeStyle = type(
    "RadarVolumeStyle", (Element,), {"name": "RadarVolumeStyle"}
)
rectangle2D = Rectangle2D = type("Rectangle2D", (Element,), {"name": "Rectangle2D"})
rectangularTorus = RectangularTorus = type(
    "RectangularTorus", (Element,), {"name": "RectangularTorus"}
)
refinementTexture = RefinementTexture = type(
    "RefinementTexture", (Element,), {"name": "RefinementTexture"}
)
remoteSelectionGroup = RemoteSelectionGroup = type(
    "RemoteSelectionGroup", (Element,), {"name": "RemoteSelectionGroup"}
)
renderedTexture = RenderedTexture = type(
    "RenderedTexture", (Element,), {"name": "RenderedTexture"}
)
rigidBody = RigidBody = type("RigidBody", (Element,), {"name": "RigidBody"})
rigidBodyCollection = RigidBodyCollection = type(
    "RigidBodyCollection", (Element,), {"name": "RigidBodyCollection"}
)
scalarChaser = ScalarChaser = type("ScalarChaser", (Element,), {"name": "ScalarChaser"})
scalarDamper = ScalarDamper = type("ScalarDamper", (Element,), {"name": "ScalarDamper"})
scalarInterpolator = ScalarInterpolator = type(
    "ScalarInterpolator", (Element,), {"name": "ScalarInterpolator"}
)
segmentedVolumeData = SegmentedVolumeData = type(
    "SegmentedVolumeData", (Element,), {"name": "SegmentedVolumeData"}
)
shadedVolumeStyle = ShadedVolumeStyle = type(
    "ShadedVolumeStyle", (Element,), {"name": "ShadedVolumeStyle"}
)
shaderPart = ShaderPart = type("ShaderPart", (Element,), {"name": "ShaderPart"})
silhouetteEnhancementVolumeStyle = SilhouetteEnhancementVolumeStyle = type(
    "SilhouetteEnhancementVolumeStyle",
    (Element,),
    {"name": "SilhouetteEnhancementVolumeStyle"},
)
singleAxisHingeJoint = SingleAxisHingeJoint = type(
    "SingleAxisHingeJoint", (Element,), {"name": "SingleAxisHingeJoint"}
)
sliderJoint = SliderJoint = type("SliderJoint", (Element,), {"name": "SliderJoint"})
slopedCylinder = SlopedCylinder = type(
    "SlopedCylinder", (Element,), {"name": "SlopedCylinder"}
)
snout = Snout = type("Snout", (Element,), {"name": "Snout"})
solidOfRevolution = SolidOfRevolution = type(
    "SolidOfRevolution", (Element,), {"name": "SolidOfRevolution"}
)
sound = Sound = type("Sound", (Element,), {"name": "Sound"})
sphereSegment = SphereSegment = type(
    "SphereSegment", (Element,), {"name": "SphereSegment"}
)
sphereSensor = SphereSensor = type("SphereSensor", (Element,), {"name": "SphereSensor"})
splinePositionInterpolator = SplinePositionInterpolator = type(
    "SplinePositionInterpolator", (Element,), {"name": "SplinePositionInterpolator"}
)
spotLight = SpotLight = type("SpotLight", (Element,), {"name": "SpotLight"})
staticGroup = StaticGroup = type("StaticGroup", (Element,), {"name": "StaticGroup"})
stippleVolumeStyle = StippleVolumeStyle = type(
    "StippleVolumeStyle", (Element,), {"name": "StippleVolumeStyle"}
)
surfaceShaderTexture = SurfaceShaderTexture = type(
    "SurfaceShaderTexture", (Element,), {"name": "SurfaceShaderTexture"}
)
switch = Switch = type("Switch", (Element,), {"name": "Switch"})
texCoordDamper2D = TexCoordDamper2D = type(
    "TexCoordDamper2D", (Element,), {"name": "TexCoordDamper2D"}
)
text = Text = type("Text", (Element,), {"name": "Text"})
texture = Texture = type("Texture", (Element,), {"name": "Texture"})
textureCoordinate = TextureCoordinate = type(
    "TextureCoordinate", (Element,), {"name": "TextureCoordinate"}
)
textureCoordinate3D = TextureCoordinate3D = type(
    "TextureCoordinate3D", (Element,), {"name": "TextureCoordinate3D"}
)
textureCoordinateGenerator = TextureCoordinateGenerator = type(
    "TextureCoordinateGenerator", (Element,), {"name": "TextureCoordinateGenerator"}
)
textureProperties = TextureProperties = type(
    "TextureProperties", (Element,), {"name": "TextureProperties"}
)
textureTransform = TextureTransform = type(
    "TextureTransform", (Element,), {"name": "TextureTransform"}
)
textureTransform3D = TextureTransform3D = type(
    "TextureTransform3D", (Element,), {"name": "TextureTransform3D"}
)
textureTransformMatrix3D = TextureTransformMatrix3D = type(
    "TextureTransformMatrix3D", (Element,), {"name": "TextureTransformMatrix3D"}
)
toneMappedVolumeStyle = ToneMappedVolumeStyle = type(
    "ToneMappedVolumeStyle", (Element,), {"name": "ToneMappedVolumeStyle"}
)
torus = Torus = type("Torus", (Element,), {"name": "Torus"})
touchSensor = TouchSensor = type("TouchSensor", (Element,), {"name": "TouchSensor"})
triangleSet = TriangleSet = type("TriangleSet", (Element,), {"name": "TriangleSet"})
triangleSet2D = TriangleSet2D = type(
    "TriangleSet2D", (Element,), {"name": "TriangleSet2D"}
)
twoSidedMaterial = TwoSidedMaterial = type(
    "TwoSidedMaterial", (Element,), {"name": "TwoSidedMaterial"}
)
uniform = Uniform = type("Uniform", (Element,), {"name": "Uniform"})
universalJoint = UniversalJoint = type(
    "UniversalJoint", (Element,), {"name": "UniversalJoint"}
)
viewfrustum = Viewfrustum = type("Viewfrustum", (Element,), {"name": "Viewfrustum"})
viewpoint = Viewpoint = type("Viewpoint", (Element,), {"name": "Viewpoint"})
volumeData = VolumeData = type("VolumeData", (Element,), {"name": "VolumeData"})
worldInfo = WorldInfo = type("WorldInfo", (Element,), {"name": "WorldInfo"})
x3DAppearanceChildNode = X3DAppearanceChildNode = type(
    "X3DAppearanceChildNode", (Element,), {"name": "X3DAppearanceChildNode"}
)
x3DAppearanceNode = X3DAppearanceNode = type(
    "X3DAppearanceNode", (Element,), {"name": "X3DAppearanceNode"}
)
x3DBackgroundNode = X3DBackgroundNode = type(
    "X3DBackgroundNode", (Element,), {"name": "X3DBackgroundNode"}
)
x3DBinaryContainerGeometryNode = X3DBinaryContainerGeometryNode = type(
    "X3DBinaryContainerGeometryNode",
    (Element,),
    {"name": "X3DBinaryContainerGeometryNode"},
)
x3DBindableNode = X3DBindableNode = type(
    "X3DBindableNode", (Element,), {"name": "X3DBindableNode"}
)
x3DBoundedObject = X3DBoundedObject = type(
    "X3DBoundedObject", (Element,), {"name": "X3DBoundedObject"}
)
x3DChaserNode = X3DChaserNode = type(
    "X3DChaserNode", (Element,), {"name": "X3DChaserNode"}
)
x3DChildNode = X3DChildNode = type("X3DChildNode", (Element,), {"name": "X3DChildNode"})
x3DColorNode = X3DColorNode = type("X3DColorNode", (Element,), {"name": "X3DColorNode"})
x3DComposableVolumeRenderStyleNode = X3DComposableVolumeRenderStyleNode = type(
    "X3DComposableVolumeRenderStyleNode",
    (Element,),
    {"name": "X3DComposableVolumeRenderStyleNode"},
)
x3DComposedGeometryNode = X3DComposedGeometryNode = type(
    "X3DComposedGeometryNode", (Element,), {"name": "X3DComposedGeometryNode"}
)
x3DCoordinateNode = X3DCoordinateNode = type(
    "X3DCoordinateNode", (Element,), {"name": "X3DCoordinateNode"}
)
x3DDamperNode = X3DDamperNode = type(
    "X3DDamperNode", (Element,), {"name": "X3DDamperNode"}
)
x3DDragSensorNode = X3DDragSensorNode = type(
    "X3DDragSensorNode", (Element,), {"name": "X3DDragSensorNode"}
)
x3DEnvironmentNode = X3DEnvironmentNode = type(
    "X3DEnvironmentNode", (Element,), {"name": "X3DEnvironmentNode"}
)
x3DEnvironmentTextureNode = X3DEnvironmentTextureNode = type(
    "X3DEnvironmentTextureNode", (Element,), {"name": "X3DEnvironmentTextureNode"}
)
x3DFogNode = X3DFogNode = type("X3DFogNode", (Element,), {"name": "X3DFogNode"})
x3DFollowerNode = X3DFollowerNode = type(
    "X3DFollowerNode", (Element,), {"name": "X3DFollowerNode"}
)
x3DFontStyleNode = X3DFontStyleNode = type(
    "X3DFontStyleNode", (Element,), {"name": "X3DFontStyleNode"}
)
x3DGeometricPropertyNode = X3DGeometricPropertyNode = type(
    "X3DGeometricPropertyNode", (Element,), {"name": "X3DGeometricPropertyNode"}
)
x3DGeometryNode = X3DGeometryNode = type(
    "X3DGeometryNode", (Element,), {"name": "X3DGeometryNode"}
)
x3DGroupingNode = X3DGroupingNode = type(
    "X3DGroupingNode", (Element,), {"name": "X3DGroupingNode"}
)
x3DInfoNode = X3DInfoNode = type("X3DInfoNode", (Element,), {"name": "X3DInfoNode"})
x3DInterpolatorNode = X3DInterpolatorNode = type(
    "X3DInterpolatorNode", (Element,), {"name": "X3DInterpolatorNode"}
)
x3DLightNode = X3DLightNode = type("X3DLightNode", (Element,), {"name": "X3DLightNode"})
x3DLODNode = X3DLODNode = type("X3DLODNode", (Element,), {"name": "X3DLODNode"})
x3DMaterialNode = X3DMaterialNode = type(
    "X3DMaterialNode", (Element,), {"name": "X3DMaterialNode"}
)
x3DMetadataObject = X3DMetadataObject = type(
    "X3DMetadataObject", (Element,), {"name": "X3DMetadataObject"}
)
x3DNavigationInfoNode = X3DNavigationInfoNode = type(
    "X3DNavigationInfoNode", (Element,), {"name": "X3DNavigationInfoNode"}
)
x3DNBodyCollidableNode = X3DNBodyCollidableNode = type(
    "X3DNBodyCollidableNode", (Element,), {"name": "X3DNBodyCollidableNode"}
)
x3DNode = X3DNode = type("X3DNode", (Element,), {"name": "X3DNode"})
x3DPlanarGeometryNode = X3DPlanarGeometryNode = type(
    "X3DPlanarGeometryNode", (Element,), {"name": "X3DPlanarGeometryNode"}
)
x3DPointingDeviceSensorNode = X3DPointingDeviceSensorNode = type(
    "X3DPointingDeviceSensorNode", (Element,), {"name": "X3DPointingDeviceSensorNode"}
)
x3DRigidJointNode = X3DRigidJointNode = type(
    "X3DRigidJointNode", (Element,), {"name": "X3DRigidJointNode"}
)
x3DSensorNode = X3DSensorNode = type(
    "X3DSensorNode", (Element,), {"name": "X3DSensorNode"}
)
x3DShaderNode = X3DShaderNode = type(
    "X3DShaderNode", (Element,), {"name": "X3DShaderNode"}
)
x3DShapeNode = X3DShapeNode = type("X3DShapeNode", (Element,), {"name": "X3DShapeNode"})
x3DSoundNode = X3DSoundNode = type("X3DSoundNode", (Element,), {"name": "X3DSoundNode"})
x3DSoundSourceNode = X3DSoundSourceNode = type(
    "X3DSoundSourceNode", (Element,), {"name": "X3DSoundSourceNode"}
)
x3DSpatialGeometryNode = X3DSpatialGeometryNode = type(
    "X3DSpatialGeometryNode", (Element,), {"name": "X3DSpatialGeometryNode"}
)
x3DTexture3DNode = X3DTexture3DNode = type(
    "X3DTexture3DNode", (Element,), {"name": "X3DTexture3DNode"}
)
x3DTextureCoordinateNode = X3DTextureCoordinateNode = type(
    "X3DTextureCoordinateNode", (Element,), {"name": "X3DTextureCoordinateNode"}
)
x3DTextureNode = X3DTextureNode = type(
    "X3DTextureNode", (Element,), {"name": "X3DTextureNode"}
)
x3DTextureTransformNode = X3DTextureTransformNode = type(
    "X3DTextureTransformNode", (Element,), {"name": "X3DTextureTransformNode"}
)
x3DTimeDependentNode = X3DTimeDependentNode = type(
    "X3DTimeDependentNode", (Element,), {"name": "X3DTimeDependentNode"}
)
x3DTouchSensorNode = X3DTouchSensorNode = type(
    "X3DTouchSensorNode", (Element,), {"name": "X3DTouchSensorNode"}
)
x3DTransformNode = X3DTransformNode = type(
    "X3DTransformNode", (Element,), {"name": "X3DTransformNode"}
)
x3DVertexAttributeNode = X3DVertexAttributeNode = type(
    "X3DVertexAttributeNode", (Element,), {"name": "X3DVertexAttributeNode"}
)
x3DViewpointNode = X3DViewpointNode = type(
    "X3DViewpointNode", (Element,), {"name": "X3DViewpointNode"}
)
x3DVolumeDataNode = X3DVolumeDataNode = type(
    "X3DVolumeDataNode", (Element,), {"name": "X3DVolumeDataNode"}
)
x3DVolumeRenderStyleNode = X3DVolumeRenderStyleNode = type(
    "X3DVolumeRenderStyleNode", (Element,), {"name": "X3DVolumeRenderStyleNode"}
)



"""
    test_webapi
    ~~~~~~~~~~~~~~~~
"""

import unittest

from domonic.dom import *
from domonic.html import *
from domonic.javascript import *
from domonic.webapi import *

# from domonic.decorators import silence


class TestCase(unittest.TestCase):

    def test_encodingAPI(self):
        # utf8decoder = TextDecoder()  # default 'utf-8' or 'utf8'

        # u8arr = Uint8Array([240, 160, 174, 183])
        # i8arr = Int8Array([-16, -96, -82, -73])
        # u16arr = Uint16Array([41200, 47022])
        # i16arr = Int16Array([-24336, -18514])
        # i32arr = Int32Array([-1213292304])

        # console.log(utf8decoder.decode(u8arr))
        # console.log(utf8decoder.decode(i8arr))
        # console.log(utf8decoder.decode(u16arr))
        # console.log(utf8decoder.decode(i16arr))
        # console.log(utf8decoder.decode(i32arr))

        # win1251decoder = TextDecoder('windows-1251')
        # b = Uint8Array([207, 240, 232, 226, 229, 242, 44, 32, 236, 232, 240, 33])
        # console.log(win1251decoder.decode(b))  # // Привет, мир!
        pass

    def test_canvas(self):
        pass

    def test_clipboard(self):
        pass

    def test_cookiestore(self):
        pass

    def test_credentialmanagement(self):
        pass

    def test_crypto(self):
        pass

    def test_cssfontloading(self):
        pass

    def test_dragndrop(self):
        pass

    def test_filereader(self):
        pass

    def test_filesysmte(self):
        pass

    def test_fetch(self):
        pass

    def test_gamepad(self):
        pass

    def test_geolocation(self):
        pass

    def test_history(self):

        from domonic.window import window

        window.location = 'https://www.google.com'
        window.location = 'https://www.facebook.com'
        window.location = 'https://www.linkedin.com'

        # print(window.history.length)
        # print(window.history.state)
        window.history.back()
        # print(window.history.state, window.location.href)
        assert window.location.href == 'https://www.facebook.com'
        # print(window.history.length)
        # print(window.history.state)
        window.history.back()
        # print(window.history.state)
        assert window.location.href == 'https://www.google.com'
        assert 'Google' in window.document.querySelector('title').text
        # print(window.history.length)
        # print(window.history.state)
        window.history.forward()
        assert window.location.href == 'https://www.facebook.com'
        assert 'Facebook' in window.document.querySelector('title').text
        # print(window.history.state)
        window.history.forward()
        # print(window.document.querySelector('title').text)
        assert 'LinkedIn' in window.document.querySelector('title').text

        print(window.history)

    def test_gl(self):
        pass

    def test_intersectobserver(self):
        pass

    def test_mediacapabilities(self):
        pass

    def test_mediasession(self):
        pass

    def test_mediastream(self):
        pass

    def test_networkinfo(self):
        pass

    def test_notifications(self):
        pass

    def test_performance(self):
        pass

    def test_permissions(self):
        pass

    def test_serversentevents(self):
        # from domonic.webapi.sse import EventSource
        # es = EventSource('/sse')
        # es.onmessage = lambda event: print(event.data)
        # es.onopen = lambda event: print('open')
        # es.onclose = lambda event: print('close')
        # es.onerror = lambda event: print('error')
        # es.send('Hello')
        
        # TESTING - useage example... clone this and point at the stream
        # https://github.com/byteface/SSELoggerDemo
        pass


    def test_xhr(self):
        from domonic.html import br, button, div, form, hr, input
        from domonic.javascript import Global
        # def on_submit(event):
        #     event.preventDefault()
        #     alert("Form submitted")
        # def on_load(event):
        #     event.preventDefault()
        #     alert("Page loaded")
        # def on_error(event):
        #     event.preventDefault()
        #     alert("Page error")
        from domonic.webapi.xhr import FormData

        myform = form(action='/', method='post')
        myform += input(type='text', name='name', placeholder='Name')
        myform += input(type='text', name='email', placeholder='Email')
        myform += input(type='text', name='phone', placeholder='Phone')
        myform += input(type='text', name='message', placeholder='Message')
        # myform += button(type='submit', value='Submit')

        f = FormData(myform)
        print(f)

        # f.append('name', 'John')
        # f.append('age', '25')
        # f.append('email', '

        print('***')

        # myform = """
        # <form action="/">
        #     <input type="text" name="name" placeholder="Name">
        #     <input type="text" name="email" placeholder="Email">
        #     <input type="text" name="phone" placeholder="Phone">
        #     <input type="submit" value="Submit">
        # </form>
        # """
        # f = FormData(myform)
        # print(f)


    def test_xpath(self):

        from domonic import domonic
        from domonic.webapi.xpath import XPathEvaluator, XPathResult

        # api unit test based on mdn example
        # https://developer.mozilla.org/en-US/docs/Web/API/XPathEvaluator

        somehtml = '''
        <div>XPath example</div>
        <div>Number of &lt;div&gt;s: <output></output></div>
        '''
        page = domonic.parseString(somehtml)  # NOTE - probably requries html5lib install
        evaluator = XPathEvaluator()
        expression = evaluator.createExpression("//div")
        result = expression.evaluate(page, XPathResult.ORDERED_NODE_SNAPSHOT_TYPE)
        assert result.snapshotLength == 2
        # print(result.nodes)

        import requests
        r = requests.get("http://eventual.technology")
        page = domonic.parseString(r.content.decode("utf-8"))

        ## Selectors

        evaluator = XPathEvaluator()
        expression = evaluator.createExpression("//h1")
        result = expression.evaluate(page, XPathResult.ORDERED_NODE_SNAPSHOT_TYPE)
        assert str(result.nodes[0]) == '<h1 class="text-uppercase hero-text text-black">We are<span class="font-weight-bold d-block">Eventual Technology</span></h1>'

        expression = evaluator.createExpression("//div//p")
        result = expression.evaluate(page, XPathResult.ORDERED_NODE_SNAPSHOT_TYPE)
        assert str(result.nodes[0]) == '<p class="headings-font-family text-uppercase lead">Welcome to the information age</p>'

        expression = evaluator.createExpression("//ul/li")
        result = expression.evaluate(page, XPathResult.ORDERED_NODE_SNAPSHOT_TYPE)
        print(str(result.nodes))

        expression = evaluator.createExpression("//ul/li/a")
        result = expression.evaluate(page, XPathResult.ORDERED_NODE_SNAPSHOT_TYPE)
        print(str(result.nodes))

        expression = evaluator.createExpression("//ul/li/a")
        result = expression.evaluate(page, XPathResult.ORDERED_NODE_SNAPSHOT_TYPE)
        print(str(result.nodes))

        expression = evaluator.createExpression("//ul/li/a")
        result = expression.evaluate(page, XPathResult.ORDERED_NODE_SNAPSHOT_TYPE)
        print(str(result.nodes))

        expression = evaluator.createExpression("//div/*")
        result = expression.evaluate(page, XPathResult.ORDERED_NODE_SNAPSHOT_TYPE)
        print(str(result.nodes))

        # root fails?
        # expression = evaluator.createExpression("/")
        # result = expression.evaluate(page, XPathResult.ORDERED_NODE_SNAPSHOT_TYPE)
        # print(str(result.nodes[0]))

        # expression = evaluator.createExpression("/body")
        # result = expression.evaluate(page, XPathResult.ORDERED_NODE_SNAPSHOT_TYPE)
        # print(str(result.nodes[0]))

        # NOTE - attributes reqiures underscores
        expression = evaluator.createExpression('//*[@_id="contact"]')
        result = expression.evaluate(page, XPathResult.ORDERED_NODE_SNAPSHOT_TYPE)
        print(str(result.nodes))

        expression = evaluator.createExpression('//*[@_class="social-link social-link-instagram"]')  # NOTE - requires all classes to match
        result = expression.evaluate(page, XPathResult.ORDERED_NODE_SNAPSHOT_TYPE)
        print(str(result.nodes))

        # expression = evaluator.createExpression("//input[@type='submit']")
        # result = expression.evaluate(page, XPathResult.ORDERED_NODE_SNAPSHOT_TYPE)
        # print(str(result.nodes))

        expression = evaluator.createExpression("//a[contains(@_href, 'twitter')]")
        result = expression.evaluate(page, XPathResult.ORDERED_NODE_SNAPSHOT_TYPE)
        print(str(result.nodes[0]))
        # wow. nice!

        expression = evaluator.createExpression("//a[contains(@href, 'twitter')]")
        result = expression.evaluate(page, XPathResult.ORDERED_NODE_SNAPSHOT_TYPE)
        print("NO UNDERSCORE", str(result.nodes[0]))

        expression = evaluator.createExpression('//*[last()][name()="a"]')
        result = expression.evaluate(page, XPathResult.ORDERED_NODE_SNAPSHOT_TYPE)
        print(str(result.nodes[0]))
        # so cool this all works out of the box!

        expression = evaluator.createExpression('//span/text()')
        result = expression.evaluate(page, XPathResult.ORDERED_NODE_SNAPSHOT_TYPE)
        print(str(result.nodes[0]))

        somepage = html(head(),
                        body(
                            h1("some title"),
                            p("some text"),
                            div("some more text")
                        )
                    )

        expression = evaluator.createExpression('//div/text()')
        result = expression.evaluate(somepage, XPathResult.ORDERED_NODE_SNAPSHOT_TYPE)
        print(str(result.nodes[0]))



        '''
        TODO - unit tests for the following so i know what works

        Descendant selectors
        h1	//h1	?
        div p	//div//p	?
        ul > li	//ul/li	?
        ul > li > a	//ul/li/a
        div > *	//div/*
        :root	/	?
        :root > body	/body

        Attribute selectors
        #id	//*[@id="id"]	?
        .class	//*[@class="class"] …kinda
        input[type="submit"]	//input[@type="submit"]
        a#abc[for="xyz"]	//a[@id="abc"][@for="xyz"]	?
        a[rel]	//a[@rel]
        a[href^='/']	//a[starts-with(@href, '/')]	?
        a[href$='pdf']	//a[ends-with(@href, '.pdf')]
        a[href*='://']	//a[contains(@href, '://')]
        a[rel~='help']	//a[contains(@rel, 'help')] …kinda

        Order selectors
        ul > li:first-of-type	//ul/li[1]	?
        ul > li:nth-of-type(2)	//ul/li[2]
        ul > li:last-of-type	//ul/li[last()]
        li#id:first-of-type	//li[1][@id="id"]	?
        a:first-child	//*[1][name()="a"]
        a:last-child	//*[last()][name()="a"]

        Siblings
        h1 ~ ul	//h1/following-sibling::ul	?
        h1 + ul	//h1/following-sibling::ul[1]
        h1 ~ #id	//h1/following-sibling::[@id="id"]

        jQuery
        $('ul > li').parent()	//ul/li/..	?
        $('li').closest('section')	//li/ancestor-or-self::section
        $('a').attr('href')	//a/@href	?
        $('span').text()	//span/text()

        Other things
        h1:not([id])	//h1[not(@id)]	?

        Text match	//button[text()="Submit"]	?
        Text match (substring)	//button[contains(text(),"Go")]

        Arithmetic	//product[@price > 2.50]

        Has children	//ul[*]
        Has children (specific)	//ul[li]
        Or logic	//a[@name or @href]	?
        Union (joins results)	//a | //div	?
        Class check
        //div[contains(concat(' ',normalize-space(@class),' '),' foobar ')]
        Xpath doesn’t have the “check if part of space-separated list” operator, so this is the workaround (source).

        #Expressions
        Steps and axes
        //	ul	/	a[@id='link']
        Axis	Step	Axis	Step
        Prefixes
        Prefix	Example	What
        //	//hr[@class='edge']	Anywhere
        ./	./a	Relative
        /	/html/body/div	Root
        Begin your expression with any of these.

        Axes
        Axis	Example	What
        /	//ul/li/a	Child
        //	//[@id="list"]//a	Descendant
        Separate your steps with /. Use two (//) if you don’t want to select direct children.

        Steps
        //div
        //div[@name='box']
        //[@id='link']
        A step may have an element name (div) and predicates ([...]). Both are optional. They can also be these other things:

        //a/text()     #=> "Go home"
        //a/@href      #=> "index.html"
        //a/*          #=> All a's child elements
        #Predicates
        Predicates
        //div[true()]
        //div[@class="head"]
        //div[@class="head"][@id="top"]
        Restricts a nodeset only if some condition is true. They can be chained.

        Operators
        # Comparison
        //a[@id = "xyz"]
        //a[@id != "xyz"]
        //a[@price > 25]
        # Logic (and/or)
        //div[@id="head" and position()=2]
        //div[(x and y) or not(z)]
        Use comparison and logic operators to make conditionals.

        Using nodes
        # Use them inside functions
        //ul[count(li) > 2]
        //ul[count(li[@class='hide']) > 0]
        # This returns `<ul>` that has a `<li>` child
        //ul[li]
        You can use nodes inside predicates.

        Indexing
        //a[1]                  # first <a>
        //a[last()]             # last <a>
        //ol/li[2]              # second <li>
        //ol/li[position()=2]   # same as above
        //ol/li[position()>1]   # :not(:first-of-type)
        Use [] with a number, or last() or position().

        Chaining order
        a[1][@href='/']
        a[@href='/'][1]
        Order is significant, these two are different.

        Nesting predicates
        //section[.//h1[@id='hi']]
        This returns <section> if it has an <h1> descendant with id='hi'.

        #Functions
        Node functions
        name()                     # //[starts-with(name(), 'h')]
        text()                     # //button[text()="Submit"]
                                # //button/text()
        lang(str)
        namespace-uri()
        count()                    # //table[count(tr)=1]
        position()                 # //ol/li[position()=2]
        Boolean functions
        not(expr)                  # button[not(starts-with(text(),"Submit"))]
        String functions
        contains()                 # font[contains(@class,"head")]
        starts-with()              # font[starts-with(@class,"head")]
        ends-with()                # font[ends-with(@class,"head")]
        concat(x,y)
        substring(str, start, len)
        substring-before("01/02", "/")  #=> 01
        substring-after("01/02", "/")   #=> 02
        translate()
        normalize-space()
        string-length()
        Type conversion
        string()
        number()
        boolean()
        #Axes
        Using axes
        //ul/li                       # ul > li
        //ul/child::li                # ul > li (same)
        //ul/following-sibling::li    # ul ~ li
        //ul/descendant-or-self::li   # ul li
        //ul/ancestor-or-self::li     # $('ul').closest('li')
        Steps of an expression are separated by /, usually used to pick child nodes. That’s not always true: you can specify a different “axis” with ::.

        //	ul	/child::	li
        Axis	Step	Axis	Step
        Child axis
        # both the same
        //ul/li/a
        //child::ul/child::li/child::a
        child:: is the default axis. This makes //a/b/c work.

        # both the same
        # this works because `child::li` is truthy, so the predicate succeeds
        //ul[li]
        //ul[child::li]
        # both the same
        //ul[count(li) > 2]
        //ul[count(child::li) > 2]
        Descendant-or-self axis
        # both the same
        //div//h4
        //div/descendant-or-self::h4
        // is short for the descendant-or-self:: axis.

        # both the same
        //ul//[last()]
        //ul/descendant-or-self::[last()]
        Other axes
        Axis	Abbrev	Notes
        ancestor
        ancestor-or-self
        attribute	@	@href is short for attribute::href
        child	div is short for child::div
        descendant
        descendant-or-self	//	// is short for /descendant-or-self::node()/
        namespace
        self	.	. is short for self::node()
        parent	..	.. is short for parent::node()
        following
        following-sibling
        preceding
        preceding-sibling
        There are other axes you can use.

        Unions
        //a | //span
        Use | to join two expressions.

        #More examples
        Examples
        //*                 # all elements
        count(//*)          # count all elements
        (//h1)[1]/text()    # text of the first h1 heading
        //li[span]          # find a <li> with an <span> inside it
                            # ...expands to //li[child::span]
        //ul/li/..          # use .. to select a parent
        Find a parent
        //section[h1[@id='section-name']]
        Finds a <section> that directly contains h1#section-name

        //section[//h1[@id='section-name']]
        Finds a <section> that contains h1#section-name. (Same as above, but uses descendant-or-self instead of child)

        Closest
        ./ancestor-or-self::[@class="box"]
        Works like jQuery’s $().closest('.box').

        Attributes
        //item[@price > 2*@discount]
        '''


if __name__ == '__main__':
    unittest.main()



"""
    domonic
    ====================================

    A library for creating html

"""

__version__ = "0.9.10"
__license__ = 'MIT'
__author__ = "@byteface"

VERSION = __version__


import re

import requests

# from domonic.components import Input
from domonic.html import *
from domonic.html import html_attributes as attributes
from domonic.javascript import *
from domonic.svg import *
from domonic.utils import Utils


class domonic:

    JS_MASTER = "assets/js/master.js"
    CSS_STYLE = "assets/css/style.css"

    @staticmethod
    def get(url: str):
        """ downloads html and converts to domonic """
        r = requests.get(url)
        return domonic.parse(r.content.decode("utf-8"))
        # TODO - param to eval

    @staticmethod
    def loads(path: str, *args, **kwargs):
        """ [
            given a path to a file will return the .pyml as a python object

            if you have variables in the template they can be pass as kwargs
        ]
        """
        with open(path, "r") as pyml_string:
            content = pyml_string.read()
            # print("++++",content, type(content) )
            prog = domonic.domonify(str(content), *args, **kwargs)
            if type(prog) is tuple:
                prog = prog[0]
            return prog

    @staticmethod
    def load(pyml: str, *args, **kwargs):
        """ [
            turns a pyml string into a python object
        ]
        """
        if not isinstance(pyml, str):
            raise ValueError("load requires a string not:", type(pyml))

        page = domonic.parse(pyml)
        prog = domonic.domonify(page, *args, **kwargs)
        if type(prog) is tuple:
            if len(prog) < 2:
                prog = prog[0]
            elif prog[1] == None:
                prog = prog[0]
        return prog

    @staticmethod
    def domonify(pyml: str, *args, **kwargs):
        """ [
            attempts to fix pyml
        ]

        Args:
            pyml (str): [a string in the form div(_class="123")]

        Returns:
            a python object
            Note:
            returns a potentially edited working program. (not the string)
            if it was ammeneded, render the returned object to get the new string
        """
        # print(pyml)
        if not isinstance(pyml, str):
            raise ValueError("domonify requires a string not:", type(pyml))

        # print("HI>>", pyml)

        s = domonic.evaluate(pyml, *args, **kwargs)

        # NOTE - valid chunks of pyml can still not eval if they are not wrapped
        # i.e. a list not in aa ul or ol. when on single line evaulate will fix
        # but on mulitple lines it will not.
        try:
            p = eval(s, {**kwargs, **globals()})
        except Exception as e:
            print("Failed to evaluate as mulitline trying again:", e)
            pyml = ''.join(pyml.splitlines()).strip(',')  # try again on a single line
            s = domonic.evaluate(pyml, *args, **kwargs)
            p = eval(s, {**kwargs, **globals()})

        return p

    LAST_ERR = None  # to stop re-eval
    @staticmethod
    def evaluate(pyml: str, *args, **kwargs):
        """ [
            attempts to fix pyml by using eval to make sure we can contruct nodes.
            be careful.
        ]

        Args:
            pyml (str): [a string in the form div(_class="123")]

        Returns:
            a python object
            Note:
            returns a potentially edited working program. (not the string)
            if it was ammeneded, render the returned object to get the new string
        """

        # print(pyml)
        if not isinstance(pyml, str):
            raise ValueError("evaluate requires a string not:", type(pyml))

        try:
            # TODO - strip any potentially bad/dangerous code before eval.
            p = eval(pyml, {**kwargs, **globals()})
            domonic.LAST_ERR = None
            return pyml  # ????
        except Exception as e:
            # import sys
            # old_log = sys.stdout
            # log_file = open("fail.log","w")
            # sys.stdout = log_file
            # print(e)
            # sys.stdout = old_log

            # if end of file err. add a closed curly
            if "EOF" in str(e):
                # unexpected EOF while parsing (<string>, line 471)
                err = str(e)
                if str(len(pyml.splitlines())) in err:
                    pyml += ")"
                    return domonic.evaluate(pyml)  # try again

            if "positional argument follows keyword argument" in str(e):

                '''
                # print(Utils.digits(e))
                if str(e) == domonic.LAST_ERR:  # only allow 1 error per line
                    # raise ValueError("Recursion limit exceeded")
                    domonic.LAST_ERR = None
                    # raise  # Exception("Recursion limit exceeded") # TODO - cant raise as called by self
                    try:
                        return
                    except Exception as e:
                        raise Exception("Recursion limit exceeded")
                else:
                    domonic.LAST_ERR = str(e)
                # return
                '''
                num = int(Utils.digits(str(e)))  # go backwards from this line. to the one before it opened
                pyml = pyml.splitlines()

                # NOTE - working backwards from the error line. we try to wrap any content.
                # if already wrapped, we don't want to wrap again. so move back 1 line until we can wrap again
                # this is because a node may take several lines.
                countback = 2
                start_line = pyml[num-countback]
                while '_' not in start_line:
                    countback += 1
                    line = pyml[num-countback]
                    if 'html' not in line:
                        start_line = line
                pyml[num-countback] = start_line + ").html("  # need to know when to close tag comma vs wrap

                # pyml[num - 2] = pyml[num - 2] + ").html(" + str(num)   # need to know when to close tag comma vs wrap
                pyml = '\n'.join(pyml)
                return domonic.evaluate(pyml)  # try again

            # TODO -  if " does not match opening parenthesis '{' (<string>, line 9)
            # TODO -  keyword argument repeated (<string>, line 617)
            # keyword argument repeated (<string>, line 3)
            # TODO - invalid syntax (<string>, line 615)
            print('Eval failed! you will have to modify the output manually')
            return pyml

    @staticmethod
    def _is_valid_pyml(line):
        """
        tests a line
        returns True or False with replacement
        """
        try:
            test_line = line.strip('\n').strip()

            if '(' in line:
                test_line = line + ')'

            if line[0] in ['"', "_", "*"]:
                test_line = "div(" + line
                if test_line[len(test_line) - 1] != ')':
                    test_line = test_line + ')'

            if line == "),":
                return True, line

            # print(test_line)
            l = eval(test_line)
            # print('PASS:', line)
            return True, line
        except Exception as e:
            print(test_line)
            print('FAIL:', line, e)
            # print(e)
            # rety fix_hyphen_tags
            if ')' in line:  # if there was a bracket return that at least
                return False, ""
            return False, ""
        return False, ""

    @staticmethod
    def dent(pyml, use_tabs=False):
        """ [
            proper dentage
        ]
        """
        tabs_or_spaces = "    "
        if use_tabs:
            tabs_or_spaces = "\t"
        dentage = 0
        lastchar = ""
        dented = ""
        for count, char in enumerate(pyml):
            if char == "(":
                dentage += 1
            if char == ")":
                dentage -= 1
            if lastchar == "\n":  # TODO - if file doesn't have newlines already
                char = tabs_or_spaces * dentage + char
            lastchar = char
            dented += char
            if dentage < 0:
                dentage = 0
        return dented

    @staticmethod
    def parse(  page: str,
                script_tags=False,
                style_tags=False,
                parse_svg=False,
                minify=False,
                # indent=True,
                remove_broken_lines=False):
        """
        HTML as input and formats to a domonic_string : the pony he comes

        the result will NOT always be valid .pyml . often params will be in wrong order.
        evaluate can be used to try and resolve param order.
        """
        if not isinstance(page, str):
            raise ValueError("Parse requires a string required not:", type(page))

        # print('parsing parsing parsing!!')

        page = ''.join(page.split('<!DOCTYPE HTML>'))
        page = ''.join(page.split('<!DOCTYPE html>'))
        page = ''.join(page.split('<!doctype html>'))
        page = ''.join(page.split('<!doctype HTML>'))

        page = ''.join(page.split('<?xml version="1.0" encoding="utf-8"?>'))
        page = ''.join(page.split('<?xml version="1.0" encoding="utf-8" ?>'))
        page = ''.join(page.split('<?xml version="1.0" encoding="UTF-8" ?>'))

        # fully strip inline css and js
        if not script_tags:
            # scripts = re.compile(r'<(script).*?</\1>(?s)')
            # TODO fix #DeprecationWarning: Flags not at the start of the expression
            scripts = re.compile(r'<(script).*?</\1>')
            page = scripts.sub('', page)

        if not style_tags:
            # css = re.compile(r'<(style).*?</\1>(?s)')
            # TODO fix #DeprecationWarning: Flags not at the start of the expression
            css = re.compile(r'<(style).*?</\1>', re.DOTALL)

            page = css.sub('', page)

        # fully strip svg and code tags
        # svg = re.compile(r'<(svg).*?</\1>(?s)')
        # page = svg.sub('', page)

        # code = re.compile(r'<(code).*?</\1>(?s)')
        # TODO fix #DeprecationWarning: Flags not at the start of the expression
        code = re.compile(r'<(code).*?</\1>')

        page = code.sub('', page)

        comments = re.compile(r'<!--(.|\s)*?-->')
        page = comments.sub('', page)
        # page = page.strip('\n').strip()

        # remove abnormal spacing between tag attributes (TODO- maybe 2 spaces is valid somewhere?)
        page = page.replace('   ', ' ')
        page = page.replace('  ', ' ')
        page = page.replace('  ', ' ')

        # special quotes
        page = page.replace('“', '&ldquo;')
        page = page.replace('”', '&rdquo;')

        # REPLACE ANY STRINGS WE MATCH ON (NOT CONTAINED IN TAGS.
        REGY = re.compile(r'(\u0028)(?![^<>]*>)')
        page = REGY.sub('$LEFTPARENTHESIS$', page)

        REGY = re.compile(r'(\u0029)(?![^<>]*>)')
        page = REGY.sub('$RIGHTPARENTHESIS$', page)

        REGY = re.compile(r'(\u005F)(?![^<>]*>)')
        page = REGY.sub('$UNDERSCORE$', page)

        REGY = re.compile(r'(U+002D)(?![^<>]*>)')
        page = REGY.sub('$HYPHEN$', page)

        REGY = re.compile(r'(U+002D)(?![^<>]*>)')
        page = REGY.sub('$QUOTE$', page)

        REGY = re.compile(r'(U+u005B)(?![^<>]*>)')
        page = REGY.sub('$LEFTSQUARE$', page)

        REGY = re.compile(r'(U+u005D)(?![^<>]*>)')
        page = REGY.sub('$RIGHTSQUARE$', page)

        REGY = re.compile(r'(U+u003D)(?![^<>]*>)')
        page = REGY.sub('$EQUALS$', page)

        REGY = re.compile(r'(U+0027)(?![^<>]*>)')
        page = REGY.sub('$SINGLEQUOTE$', page)

        def encode_content(content=""):
            content = content.replace(")", "$RIGHTPARENTHESIS$")
            content = content.replace("(", "$LEFTPARENTHESIS$")
            content = content.replace("_", "$UNDERSCORE$")
            content = content.replace("-", "$HYPHEN$")
            content = content.replace('"', "$QUOTE$")
            content = content.replace('[', "$LEFTSQUARE$")
            content = content.replace(']', "$RIGHTSQUARE$")
            content = content.replace('=', "$EQUALS$")
            content = content.replace("'", "$SINGLEQUOTE$")
            # content = content.replace(" ", "$SPACE$")
            return content

        def encode_attr_content(content=""):
            content = content.replace(")", "$RIGHTPARENTHESIS$")
            content = content.replace("(", "$LEFTPARENTHESIS$")
            content = content.replace("_", "$UNDERSCORE$")
            content = content.replace("-", "$HYPHEN$")
            content = content.replace('"', "$QUOTE$")
            content = content.replace('[', "$LEFTSQUARE$")
            content = content.replace(']', "$RIGHTSQUARE$")
            content = content.replace('=', "$EQUALS$")
            content = content.replace("'", "$SINGLEQUOTE$")

            content = content.replace(",", "$COMMA$")
            content = content.replace(";", "$SEMICOLON$")

            return content


# LEFT_CURLY_BRACKET = \u007B
# OR = \u007C
# RIGHT_CURLY_BRACKET = \u007D
# TILDE = \u007E
# SPACE = \u0020

# [ # \u005B
# \ # \u005C
# ] # \u005D
# ^ # \u005E
# _ # \u005F
# ` # \u0060
# : # U+003A
# ; # U+003B
# < # U+003C
# = # U+003D
# > # U+003E
# ? # U+003F
# @ # U+0040
# SPACE = \u0020
# ! # U+0021
# " # U+0022
# # # U+0023
# $ # U+0024
# % # U+0025
# & # U+0026
# ' # U+0027
# ( # U+0028
# ) # U+0029
# * # U+002A
# + # U+002B
# , # U+002C
# - # U+002D
# . # U+002E
# / # U+002F

        tags = html_tags + svg_tags
        tags.sort(key=len, reverse=True)
        for tag in tags:
            page = re.sub(f"<{tag}>", f'\n{tag}(\n', page, flags=re.IGNORECASE)
            page = re.sub(f"<{tag} ", f'\n{tag}(\n', page, flags=re.IGNORECASE)
            page = re.sub(f"</{tag}>", '\n),\n', page, flags=re.IGNORECASE)

            reg = '/>'  # NOTE - er?? this is global!
            pattern = re.compile(reg)
            page = re.sub(pattern, '\n),\n', page)  # , flags=re.IGNORECASE )

        # close any tags that aren't properly self closing
        flag = False
        open_count = 0
        increase_index = 0  # by the amount of new chars you add
        last_tag = None
        for index, char in enumerate(page):
            index = index + increase_index  # TODO does this need to go back to zero.? is any of this code still relevant?
            if char == "(":
                open_count += 1
                flag = (open_count > 0)
                tag = page[index - 4] + page[index - 3] + page[index - 2] + page[index - 1]
                last_tag = tag
            if char == ")":
                open_count -= 1
                flag = (open_count > 0)
            if char == ">":# and flag is True:
                if 'meta' in tag or 'link' in tag or 'hr' in tag:  # ??... dont think this is catching anymore
                    page = f"{page[:index]}\n),\n{page[index+1:]}"
                    increase_index += 3
                    open_count -= 1
                    flag = (open_count > 0)
                    continue
                page = f'{page[:index]},\n{page[index+1:]}'
                increase_index += 1

        attribs = attributes
        attribs.append('as')
        attribs.append('prefix')
        attribs.append('role')
        attribs.append('decoding')
        attribs.append('typography')
        attribs.append('content')

        solo_attributes = [  # ones that can have no value
            "allowfullscreen", "allowpaymentrequest", "async", "autofocus", "autoplay", "checked", "controls", "default",
            "disabled", "formnovalidate", "hidden", "ismap", "itemscope", "loop", "multiple", "muted", "nomodule", "novalidate",
            "open", "playsinline", "readonly", "required", "reversed", "selected", "truespeed", "typemustmatch", "compact",
            "nohref", "noresize", "noshade", "nowrap", "scrolling", "seamless", "sortable", "autocomplete", "border", "challenge",
            "keyparams", "keygen", "spellcheck", "translate", "indeterminate"]

        solo_attributes.append('mozdisallowselectionprint')
        solo_attributes.append('moznomarginboxes')
        solo_attributes.append('crossorigin')

        # adds a comma before special attribute types
        extras = ['data-', 'aria-', 'accept-charset', "http-"]
        for attr in extras:
            reg = f' {attr}'
            pattern = re.compile(reg)
            page = re.sub(pattern, f', {attr}', page)

        # put underscores on all the attr
        for attr in attribs:
            page = re.sub(f' {attr}="', f' _{attr}="', page, flags=re.IGNORECASE)

        # commas between them
        # for attr in attribs:
            # reg = f' _{attr}="'
            # pattern = re.compile(reg)
            # page = re.sub(pattern, f' _{attr}="', page)

        # TODO - diff between loaded and inline
        # TODO - would have to replace all tags in js (same as content ) (or do opposite way round)
        # get the style and script tags
        #// sure this doesnt' work anymore as we do all tags already?
        htmltags = ["style", "script"]
        for tag in htmltags:
            reg = f"<{tag}>"
            pattern = re.compile(reg)
            page = re.sub(pattern, f'{tag}("""', page)  # , flags=re.IGNORECASE )

            # second pass. atrributed
            reg = f"<{tag}"
            pattern = re.compile(reg)
            page = re.sub(pattern, f'{tag}("""', page)  # , flags=re.IGNORECASE )

            reg = f"</{tag}>"
            pattern = re.compile(reg)
            page = re.sub(pattern, '"""),', page)  # , flags=re.IGNORECASE )

            # reg = '/>'
            # pattern = re.compile(reg)
            # page = re.sub(pattern, '"""),', page)  # , flags=re.IGNORECASE )

        page = '\n)\n'.join(page.split(',)'))  # newline this one?. not sure about this one anymore. seems brutal at this stage

        customtags = re.findall(r'<[-a-zA-Z]+', page)
        if len(customtags) > 0:
            for t in customtags:
                # print(t)
                page = page.replace(t, '\ncreate_element(\n"' + t.lstrip('<') + '"')  # < note. changed to not closing tag

        customtags = re.findall(r'<[/][-a-zA-Z]+', page)
        if len(customtags) > 0:
            for t in customtags:
                # print(t)
                page = page.replace(t, '\n),\n')  # < note. changed to not closing tag

        # any stragglers or custom tags
        page = page.replace('/>', '\n),\n')
        page = page.replace('>', '\n(\n')
        page = page.replace('<', '')

        # print(":::",page)
        # page = page.replace('>', '\n,\n')  # < note. changed to not closing tag
        # page = page.replace('<', '\n(\n')

        # < -------- END OF FIRST PASS

        def fix_hyphen_tags(line):
            # NOTE - bad! as will skip some params
            # if line.count('"') % 2 == 1:
                # return line # its an opening multi-line string so continue.

            values = re.findall('"([^"]*)"', line)
            if len(values) > 0:
                for value in values:
                    line = line.replace(value, encode_attr_content(value))

            values = re.findall("'([^']*)'", line)
            if len(values) > 0:
                for value in values:
                    line = line.replace(value, encode_attr_content(value))

            params = line.replace('" _', '", _')
            params = line.strip().strip(',').strip().split(',')

            for count, each in enumerate(params):
                parts = each.split('="')
                if len(parts) < 2:
                    continue
                key = parts[0].strip()
                val = parts[1].strip()

                if 'style' in key or 'title' in key:
                    for i, att in enumerate(attributes):
                        val = val.replace(att, '$DoMo' + str(i) + 'NiC$')

                # checks string lines have quotes both sides
                if val is None or val == "":
                    val = 'true'
                if val == ' ':
                    val = '" "'
                if val[-1] not in ['"', ",", "*", ")", "$QUOTE$"]:
                    val = val + '"'
                if val[0] not in ['"', ",", "*", "("]:  # note. added opener.
                    val = '"' + val
                if val == None or val == '"':
                    val = '""'

                val = val.replace("-", "$HYPHEN$")

                if '_' in key:  # or '_' not in key: # skip as its a single attribute with multiple key:values
                    if '-' not in key:  # TODO - may still have to do other ones as below?
                        newparam = f'{key}={val}'
                        params[count] = newparam
                        continue

                if '-' in key:
                    key = key.lstrip('_')  # if already has an underscore remove it as we add it below
                    END = ""
                    if len(line) - (line.find(val) + len(val)) < 3:
                        # print('last attribute in line')
                        END = ","

                    newparam = f'**\u007b"_{key}":{val}\u007d{END}'
                    params[count] = newparam
                elif '_' not in key and '-' not in key:  # i dont think much gets to here then anymore?
                    newparam = f', _{key}={val}'
                    params[count] = newparam

            line = ', '.join(params)
            line = line.replace('" _', '", _')

            if(line[len(line) - 1] in ["'", '"', ")", "$QUOTE$", "}", "e"]):  # TODO 'e' is the last letter of True. crap check
                line = line + ","

            return line

        # def replace_between(line, match, replacement, start=0, end=0):
        #     front = line[0:start]
        #     mid = line[start:end]
        #     end = line[end:len(line)]
        #     mid = mid.replace(match, replacement)
        #     newline = front + mid + end
        #     return newline

        def parse_attributes(line):

            values = re.findall('"([^"]*)"', line)
            if len(values) > 0:
                for value in values:
                    line = line.replace(value, encode_attr_content(value))

            # import re
            # values = re.findall("'([^']*)'", line)
            # if len(values)>0:
            #     for value in values:
            #         line = line.replace(value,encode_attr_content(value))

            # NOTE - bad! as will skip some params
            if line.count('"') % 2 == 1:
                return line  # its an opening multi-line string so continue.
                # continue

            # prevents single attrib with missing quotes from losing content. (or throw unclean html errors?)(start doing that you'll never stop)
            if line.count('=') < 2:  # if only 1 attr
                if line.count('"') < 1 and line.count("'") < 1:
                    line = line.replace(',', '$COMMA$')
                    parts = line.split('=')
                    line = parts[0] + "=" + '"' + parts[1] + '"'
                    # print(line)

            attribs = re.findall(r"((?:(?!\s|=).)*)\s*?=\s*?[\"']?((?:(?<=\")(?:(?<=\\)\"|[^\"])*|(?<=')(?:(?<=\\)'|[^'])*)|(?:(?!\"|')(?:(?!\/>|>|\s).)+))", line)

            if attribs:
                for each in attribs:
                    key = each[0].strip()
                    val = each[1].strip()
                    oldval = val

                    is_quote = lambda x: x == '"' or x == "'"
                    try:
                        has_left_quote = is_quote(line[line.find(val) - 1])
                    except Exception as e:
                        has_right_quote = False

                    try:
                        has_right_quote = is_quote(line[line.find(val) + (len(val))])
                    except Exception as e:
                        has_right_quote = False

                    val = val.replace('"', '&quot;')  # they don't always get caught by encode
                    val = val.replace(',', '&#44;')
                    # val = val.replace(';',' &#59;')

                    if 'style' in key or 'title' in key:
                        for i, att in enumerate(attributes):
                            val = val.replace(att, '$DoMo' + str(i) + 'NiC$')

                    if '-' in key and key[0] != '_':
                        line = line.replace(key, '_' + key)
                        continue

                    if 'font-size' in key:
                        line = line.replace('font-size', key)  # update to prepended underscore
                        continue  # these keys are transformed later

                    # val = val.replace("\n", "") # remove newlines in atttribute content as causes EOL when parsing

                    if len(key) > 20 or '//' in key:  # data-analytics-exit-link << NOTE 15 limit easilty buckles. bad way to check for content in keys
                        continue

                    # if key not in attributes: # THEN IT MUST BE NORMAL TEXT. strict tho
                    # continue
                    if key.istitle():  # very weak check for normal text TODO. normal text with equals gets through.
                        continue

                    if val == None or val == "":
                        val = 'true'

                    newval = ""
                    if type(val) != bool:

                        # checks string lines have quotes both sides
                        if val == None:
                            val = '""'
                        if val == ' ':
                            val = '" "'
                        if not has_right_quote:
                            val = val + '"'
                        if not has_left_quote:
                            val = '"' + val
                        if val == None or val == '"':
                            val = '""'

                        newval = val.replace("-", "$HYPHEN$")

                    if '_' not in key and '-' not in key:
                        if ':' in key:  # i.e. xml:"lang=en-US"
                            parts = key.split(':')
                            key = parts[0]
                            if len(parts) > 1:
                                if len(parts[1]) > 1:
                                    newval = parts[1] + "=" + newval
                                    line = line.replace(":" + parts[1], "")

                        line = line.replace(key + "=", ', _' + key + "=")
                        line = line.replace(key + " =", ', _' + key + "=")
                        if type(val) != bool and len(val) > 0:
                            line = Utils.replace_between(line, oldval, str(newval), line.find(key), line.find(key) + (len(key) - 1) + (len(val) - 1) + 1)  # final +1 is the equal sign

            line = line.replace('" _', '", _')
            line = line.replace("' _", "', _")  # single quote version of same thing
            # line = line.strip()
            return line

        # SECOND PASS. split onto lines and fix hyphen tags
        cleaned = []
        lines = page.splitlines()

        if 'doctype' in lines[0].lower():
            lines[0] = ""

        lines_iterator = enumerate(lines)
        for count, line in lines_iterator:
            line = line.strip()
            line = line.replace('    ', ' ')
            line = line.replace('   ', ' ')
            line = line.replace('  ', ' ')
            line = line.replace('  ', ' ')

            if len(line) < 1:
                continue

            if line == ",":
                if len(lines[count - 1]) > 0:
                    if lines[count - 1][len(lines[count - 1]) - 1] == ',':
                        continue

            if '=' in line:
                line = parse_attributes(line)  # < TODO -  normal content with equals in is getting caught here

                # solo attributes

                # TODO - should really be doing these much sooner no?
                # TODO - breaking class in css content when they have attribute names .i.e. hidden. SORTDE> shoudl be fixed now
                # aria-hidden also affected.?. by why it doing with no spaces
                if '(' not in line and ')' not in line and line[0] != '"':  # TODO - not if it already has an equals
                    for each in solo_attributes:
                        pos = line.find(each)
                        # if pos < 1: continue

                        # if the previous attribute has a leading quote already don't prepend one
                        # we assume it doesn't to start.
                        has_leading_quote = False
                        PREP = '"'
                        if pos > 5:
                            # check 4 chars back if quote set false.
                            if '"' in line[pos - 5:pos]:  # TODO - or if just the word True
                                has_leading_quote = True
                            if has_leading_quote:
                                PREP = ''

                        # if solo is first in the line
                        reg = f'^{each} (?=(?:[^"]*"[^"]*")*[^"]*$)'  # space in front. nothing behind
                        pattern = re.compile(reg)
                        line = re.sub(pattern, f'_{each}=True,', line)

                        reg = f' {each}(?=(?:[^"]*"[^"]*")*[^"]*$)'  # space in front. nothing behind
                        pattern = re.compile(reg)
                        line = re.sub(pattern, f'{PREP}, _{each}=True,', line)

                        reg = f'{each},(?=(?:[^"]*"[^"]*")*[^"]*$)'  # with a trailing comma
                        pattern = re.compile(reg)
                        line = re.sub(pattern, f'{PREP}, _{each}=True,', line)

                        reg = f' {each} (?=(?:[^"]*"[^"]*")*[^"]*$)'  # with a trailing space
                        pattern = re.compile(reg)
                        line = re.sub(pattern, f'{PREP}, _{each}=True,', line)

                        reg = f',{each} (?=(?:[^"]*"[^"]*")*[^"]*$)'  # leading comma, with a trailing space
                        pattern = re.compile(reg)
                        line = re.sub(pattern, f'{PREP}, _{each}=True,', line)

                # TODO - custom solo attributes

                line = fix_hyphen_tags(line)

                # any leftover solo hyphenataed data-tags
                hyphenated = re.findall(r' [-A-Za-z]+\w+(?:-\w+)+', line)
                for each in hyphenated:
                    line = line.replace(each, f'**\u007b"_{each}":{True}\u007d,')

            # TODO - some attribute content could have open curlies. need to replace all normal text chars
            if '(' not in line[0:10]:
                if ')' not in line[0:2]:
                    # normal text could start with underscore. so could also check for =
                    if '_' not in line:
                        if '-' in line and '=' in line:
                            # its probably a line with hypened a data-tags
                            # line = fix_hyphen_tags(line)
                            pass
                        else:
                            # its regular text content
                            line = encode_content(line)
                            line = f'"{line}"'

            is_multiline_string = False
            if line.count('"') % 2 == 1:  # find opening quotes to multilines (odd number)

                if count < len(lines) - 1:
                    next_line = lines[count + 1]
                if count > 0:
                    prev_line = lines[count - 1]

                # if its just a class and not content. bring them up onto the same line
                if '_class' in line:
                    if '(' not in next_line:  # and '"' not in next_line:
                        line = line + lines.pop(count + 1)  # merge the next line to this one
                        line = line.replace('\n', "")
                        line = line.replace("  ", " ")
                        if line.count('"') % 2 == 1:  # if still odd
                            line = line + '"'  # add a quote
                        next(lines_iterator, None)  # skip the iterator along by 1

                else:
                    if not is_multiline_string == False:
                        x = line.rindex('"')
                        if x:
                            line = line[:x] + '"""' + line[x + 1:]
                            is_multiline_string = True
                    else:
                        x = line.find('"')
                        if x:
                            line = line[:x] + '""",' + line[x + 1:]
                            is_multiline_string = False

            cleaned.append(line)
        page = '\n'.join(cleaned)

        # a final pass to try check for missing commas between lines by checking 1 line ahead
        fixed = []
        lines = page.splitlines()
        for count, line in enumerate(page.splitlines()):
            line = line.strip('\n')
            line = line.strip()
            if line[len(line) - 1] == '"':
                if count < len(lines) - 1:
                    if lines[count + 1][0] != ')':
                        line = line + ','
            if "_" in line:  # normal text can have underscores. this will break
                line = fix_hyphen_tags(line)

            if len(line) < 5 and '"' in line:  # need to stop making these in first place
                if line == '",",':
                    continue

            fixed.append(line)
        page = '\n'.join(fixed)

        def clean_junk(page):
            page = page.replace('",","', '","')
            page = page.replace('",",', '",')
            page = page.replace('", ",', '",')
            page = page.replace('", \n",', '",')
            page = page.replace('",\n ",', '",')
            page = page.replace('","\n)', '"\n)')
            page = page.replace('", \n, _', '",\n_')
            page = page.replace(', "\n)', '\n)')

            # page = page.replace(',",', ',') < VALID
            page = page.replace(',  ,', ',')  # < new bug. due to single attributes having big space in front for some reason
            page = page.replace(', ",', ',')
            page = page.replace(',,', ',')
            page = page.replace(', ,', ',')
            page = page.replace(',"",', ',')  # careul. new and covers up somethings else. solo attributes still not done well

            page = page.replace('( ,*', '(*')
            page = page.replace('( , *', '(*')
            page = page.replace('(,*', '(*')
            page = page.replace('(, *', '(*')
            page = page.replace('(,  *', '(*')
            page = page.replace('(", *', '(*')
            page = page.replace('(\n",', '(\n')
            page = page.replace('(\n,', '(')

            page = page.replace('),",', '),')
            page = page.replace('),\n"\n),', '),\n),')
            page = page.replace('},\n"\n),', '}\n),')

            page = page.replace('"_, _', '"_')  # when solo hyphenated custom attribute is first on a line.

            # page = page.replace('),\n",\n', '(')  # < break things but is also valid. text sentences can start with a comma
            # 2 issues. this also turns a closer into an opener. when catching a true case

            return page

        page = clean_junk(page)
        # page = clean_junk(page)
        # clean_junk(page)

        # put content text back to normal
        page = page.replace("$RIGHTPARENTHESIS$", ")")
        page = page.replace("$LEFTPARENTHESIS$", "(")
        page = page.replace("$UNDERSCORE$", "_")
        page = page.replace("$HYPHEN$", "-")
        page = page.replace("$QUOTE$", '&quot;')
        page = page.replace("$LEFTSQUARE$", '[')
        page = page.replace("$RIGHTSQUARE$", ']')
        page = page.replace("$EQUALS$", '=')
        page = page.replace("$SINGLEQUOTE$", "'")
        page = page.replace("$COMMA$", ",")
        page = page.replace("$SEMICOLON$", ";")

        for count, att in enumerate(attributes):
            page = page.replace('$DoMo' + str(count) + 'NiC$', att)  # undo encoding that saves attr content

        if remove_broken_lines:
            print('attempting to remove broken lines')
            fixed = []
            for count, line in enumerate(page.splitlines()):
                line = line.strip('\n')
                is_fixed, newline = domonic._is_valid_pyml(line)
                if is_fixed:
                    fixed.append(newline)
                else:  # break line into bits to keep any working parts
                    # print("BAD:", line)
                    parts = line.split(',')
                    keepers = []
                    for piece in parts:
                        piece = piece.strip()
                        piece = piece.strip('\n')
                        is_working, p = domonic._is_valid_pyml(piece)
                        # print(is_working,p)
                        if is_working:
                            keepers.append(p)
                    line = ','.join(keepers)
                    is_fixed, newline = domonic._is_valid_pyml(line + ",")
                    if is_fixed:
                        fixed.append(newline)
                        # print("FIXED:", line)
            page = '\n'.join(fixed)

        # page = ''.join(page.splitlines())
        # page = ''.join(page.splitlines())

        # if not minify and indent:
        #     print('>>',len(page))
        # page = domonic.dent(page)
        #     print('<<',len(page))

        return page

    parseString_prev_error = None
    @staticmethod
    def parseString(string):  #, parser=None):
        """ Parse a file into a DOM from a string. """
        # if parser is None:
        # import xml.parsers.expat.ExpatError as ExpatError        # parser = xml.parsers.expat.ParserCreate()

        # TODO - this needs to be off for debugging
        try:
            # if users has html5lib installed used that to parse
            import html5lib
            if 'html5lib' in sys.modules:
                from html5lib import HTMLParser

                from domonic.ext.html5lib_ import getTreeBuilder
                parser = HTMLParser(tree=getTreeBuilder())
                page = parser.parse(string)
                # print('PARSED WITH HTML5 LIB')
                return page
        except ImportError:
            pass

        try:
            from domonic.parsers import expatbuilder
            return expatbuilder.parseString(string)
        except Exception as e:
            # TODO - problem with this method. is it takes literally forever.
            # as it removes 1 char then reparses entire doc. even on small pages this is a problem.
            print(e)
            dodgycharIndex = int(Utils.digits(str(e).split(',')[1]))
            # string[int(dodgycharIndex)-1] = Utils.escape(string[int(dodgycharIndex)-1])
            dodgyChar = string[int(dodgycharIndex) - 1]
            string = Utils.replace_between(string, dodgyChar, '', dodgycharIndex-2, dodgycharIndex+2)
            if domonic.parseString_prev_error != dodgycharIndex:
                domonic.parseString_prev_error = dodgycharIndex
                return domonic.parseString(string)
            else:
                print('failed to parse invalid xml. clean and try again', e)

        # else:
            # from xml.dom import pulldom
            # return _do_pulldom_parse(pulldom.parseString, (string,),
                                    # {'parser': parser})



"""
    domonic.geom.shape.circle
    ====================================
    written by.ai
"""

from domonic.geom.shape import Shape
from domonic.javascript import Math


class Circle(Shape):
    def __init__(self, x: float, y: float, radius: float = 1.0, color=None) -> None:
        """[Circle(x, y, radius=1.0, color=None)]

        Args:
            x ([float]): [the x coordinate of the center of the circle]
            y ([float]): [the y coordinate of the center of the circle]
            radius (float, optional): [description]. Defaults to 1.0.
            color ([type], optional): [description]. Defaults to None.
        """
        super().__init__(color)
        self.radius = radius  # Create an instance variable radius

    @property
    def area(self) -> float:
        """[area]

        Returns:
            [float]: [the area of the circle]
        """
        return self.radius * self.radius * Math.PI

    @property
    def perimeter(self) -> float:
        """[perimeter]

        Returns:
            [float]: [the perimeter of the circle]
        """
        return 2 * self.radius * Math.PI

    @property
    def average_circumference(self):
        return 2 * self.radius

    @property
    def center(self):
        x, y = self.center = [self.radius, self.radius]
        return x, y

    @center.setter
    def center(self, center):
        self._center = center

    def __str__(self):
        return f"Circle({self.center}, {self.radius}, {self.color})"

    def __getstate__(self):
        return {"center": self.center, "radius": self.radius}

    def __setstate__(self, state):
        self.center = state["center"]
        self.radius = state["radius"]

    def __copy__(self):
        return Circle(self.center, self.radius, self.color)

    def __deepcopy__(self, memo):
        return Circle(self.center, self.radius, self.color)

    def __contains__(self, other):
        return other in self.center

    def __len__(self):
        return len(self.center)

    def __getitem__(self, key):
        return self.center[key]

    def __setitem__(self, key, value):
        self.center[key] = value

    def __iter__(self):
        return iter(self.center)

    def __add__(self, other):
        return Circle(self.center + other.center, self.radius + other.radius)

    def __sub__(self, other):
        return Circle(self.center - other.center, self.radius - other.radius)

    def __mul__(self, other):
        return Circle(self.center * other, self.radius * other)

    # def __rmul__(self, other):
    #     return Circle(self.center * other, self.radius * other


class Oval(Circle):
    def __init__(self, radius=2.5, size=3):
        super().__init__(size, "green")
        self.radius = radius
        self.x = self.width / 2 - self.radius / 2
        self.y = self.width / 2 + self.radius / 2


class Ellipse(Shape):
    def __init__(self, x, y, width, height, color=None):
        super().__init__(color)
        self.x = x
        self.y = y
        self.width = width
        self.height = height



"""
    domonic.sitemap
    ====================================

    generate or load sitemaps

    warning - when using image and video tags from this package they will be namespaced i.e <image:image> and <video:video>
    so i'd advise to only import them within the def that you use them in to avoid conflict with html.image

"""

import datetime

from domonic.dom import Document, Element

# __all__ = ['sitemap', 'url', 'lastmod']

sitemap_tags = [
    "sitemapindex",
    "sitemap",
    "urlset",
    "url",
    "loc",
    "lastmod",
    "changefreq",
    "priority",
    "image:image",
    "image:loc",
]

sitemap_attributes = [
    "xmlns",
    "xmlns:xsi",
    "xsi:schemaLocation",
    "xmlns:xhtml",
    "xmlns:xlink",
    "xmlns:atom",
    "xmlns:geo",
]
# sitemap_change_frequencies = ['always', 'hourly', 'daily', 'weekly', 'monthly', 'yearly', 'never']

XMLNS: str = "http://www.sitemaps.org/schemas/sitemap/0.9"
XMLNS_XSI: str = "http://www.w3.org/2001/XMLSchema-instance"
SCHEMA_SITEINDEX: str = "http://www.sitemaps.org/schemas/sitemap/0.9/siteindex.xsd"
SCHEMA_SITEMAP: str = "http://www.sitemaps.org/schemas/sitemap/0.9/sitemap.xsd"
#     xmlns="http://www.google.com/schemas/sitemap-image/1.1"
#     xmlns:video="http://www.google.com/schemas/sitemap-video/1.1">


def sitemap_format(self, *args, **kwargs):
    """attempts to prettify the output of the sitemap."""
    outp = f"<{self.name}{self.__attributes__}>{self.content}</{self.name}>"
    from domonic.xml import prettify

    return prettify(outp)


sitemapindex = type(
    "sitemapindex",
    (Document,),
    {
        "name": "sitemapindex",
        "xmlns": XMLNS,
        "xmlns:xsi": XMLNS_XSI,
        "xsi:schemaLocation": SCHEMA_SITEINDEX,
        "__format__": sitemap_format,
    },
)

sitemap = type("sitemap", (Element,), {"name": "sitemap"})

urlset = type(
    "urlset",
    (Element,),
    {
        "name": "urlset",
        "xmlns:xsi": XMLNS_XSI,
        "xsi:schemaLocation": SCHEMA_SITEMAP,
        "xmlns": XMLNS,
    },
)


url = type("url", (Element,), {"name": "url"})
loc = type("loc", (Element,), {"name": "loc"})
lastmod = type("lastmod", (Element,), {"name": "lastmod"})
changefreq = type("changefreq", (Element,), {"name": "changefreq"})
priority = type("priority", (Element,), {"name": "priority"})


def sitemapindex_from_urls(urls):
    """
    Create a sitemap index from a list of urls.

    WARNING:
    there's a difference between a sitemap index and a sitemap.
    make sure you know what you want.

    # i.e

    # <?xml version="1.0" encoding="UTF-8"?>
    # <sitemapindex xmlns="http://www.sitemaps.org/schemas/sitemap/0.9">
    # 	<sitemap>
    # 		<loc>https://xyz.com/sitemap1.xml</loc>
    # 		<lastmod>2021-07-08T13:12:16+00:00</lastmod>
    # 	</sitemap>
    # </sitemapindex>

    """
    sitemap_index = sitemapindex()
    for url in urls:
        sitemap_index.append(sitemap(loc(url), lastmod(datetime.datetime.now())))


def sitemap_from_urls(urls):
    """
    Create a sitemap from a list of urls.add()

    Note: This won't allow you to add priority or changefreq of the urls. or add images etc
    tho u could loop the nodes afterwards and do that.

    WARNING:
    there's a difference between a sitemap index and a sitemap.
    make sure you know what you want.

    """
    sitemap = urlset()
    for url in urls:
        sitemap.append(url(loc(url), lastmod(datetime.datetime.now())))
    return sitemap


def get_sitemap(path: str, *args, **kwargs):
    # use requests to downlaod a sitemap
    import requests

    r = requests.get(path)
    import domonic

    some_sitemap = domonic.domonic.parseString(r.text)
    return some_sitemap


globals()["image:image"] = type(
    "image:image", (Element,), {"name": "image", "ns": "image"}
)

globals()["image:loc"] = type("image:loc", (Element,), {"name": "loc", "ns": "image"})



"""
    domonic.webapi.webstorage
    ====================================
    https://developer.mozilla.org/en-US/docs/Web/API/Storage
    
    TODO - add more than just json as options.
    TODO - i dont believe there's any unit tests for this.
    
"""

import json
import os

from domonic.events import StorageEvent


class Storage():

    def __init__(self, filepath: str=None) -> None:
        """[localstorage. destroys on each session unless you pass the optional filepath]

        Args:
            filepath ([str], optional): [filepath]. give us a file to write to
        """
        self.storage = {}
        self.has_file = False
        if filepath:
            self.filepath = filepath
            self.has_file = True
        # check if file exists. if so load it in . if not create it
        if filepath:
            if os.path.exists(filepath):
                with open(filepath, 'r') as f:
                    self.storage = json.load(f)
            else:
                with open(filepath, 'w') as f:
                    json.dump(self.storage, f)

    def __getitem__(self, key: str) -> str:
        return self.storage[key]
    getItem = __getitem__

    def __setitem__(self, key: str, value: str) -> None:
        self.storage[key] = value
        if self.has_file:
            self._save()
    setItem = __setitem__
    
    def __getattr__(self, key: str) -> str:
        return self.storage.get(key, None)
    
    def __setattr__(self, key: str, value: str) -> None:
        if key == 'storage':
            self.__dict__[key] = value  # TODO - tests required. not sure this is correct. shouldn't it be self.storage = value?
        else:
            self.storage[key] = value
            self._save()

    def __len__(self) -> int:
        return len(self.storage.keys())

    @property
    def length(self) -> int:
        """ Returns an integer representing the number of data items stored in the Storage object. """
        return len(self)

    def _save(self) -> None:
        if self.has_file:
            with open(self.filepath, 'w') as f:
                json.dump(self.storage, f)
            return True
        return False

    def key(self, keyName: str) -> str:
        """[returns the value of the key or None if the key does not exist]

        Args:
            keyName (str): [key to get]

        Returns:
            str: [the key or None]
        """
        return self.storage.get(keyName, None)

    def removeItem(self, keyName: str) -> None:
        """[removes the key and its value from the storage]

        Args:
            keyName (str): [key to remove]
        """
        if keyName in self.storage:
            del self.storage[keyName]
            self._save()

    def clear(self) -> None:
        """ Removes all items from the storage """
        self.storage = {}
        self._save()



"""
    test_components
    ~~~~~~~~~~~~
    tests for components
"""

import unittest

from domonic.html import *
from domonic.JSON import *

# import json
# import requests
# from mock import patch



# class TestCase(unittest.TestCase):

#     def test_json(self):

# if __name__ == '__main__':
#     unittest.main()



"""
    domonic.i18n.zu
    ====================================

    html tags in Zulu

    first translated to Zulu using google translate.
    feel free to send a pull request if you think it should be different
"""
from domonic.html import *

umzimba = body
ikhanda = head
iskripthi = script
isitayela = style
# portal = portal
# strong = strong
# blockquote = blockquote
ithebula = table
ifomu = form
ilebula = label
hambisa = submit
isihloko = title
umbhalo = noscript
isigaba = section
# nav = nav
isihloko = article
eceleni = aside
iqembu = hgroup
ikheli = address
umfanekiso = figure
# figcaption = figcaption
encane = small
caphuna = cite
ikhodi = code
# sub = sub
# sup = sup
umaka = mark
irubi = ruby
# span = span
# iframe = iframe
ividiyo = video
umsindo = audio
indwangu = canvas
amazwibela = caption
iqembu = colgroup
umzimba = tbody
inqola = thead
izinyawo = tfoot
insimu = fieldset
inganekwane = legend
inkinobho = button
khetha = select
uhlulwedatha = datalist
iqembu = optgroup
# option = option
indawo = textarea
okukhiphayo = output
inqubekelaphambili = progress
imitha = meter
imininingwane = details
isifinyezo = summary
imenyu = menu
# imenyu = menuitem
ifonti = font
unhlokweni = header
unyaweni = footer
isisekelo = base
isixhumanisi = link
metha = meta
# img = img
shumeka = embed
umthombo = source
ithrekhi = track
indawo = area
ikholo = col
okokufaka = input
umyalo = command
# main = main
isisekelo = basefont
isikhungo = center
shumeka = embed
# isindex = isindex
uhlu = listing
umbhaloongenalutho = plaintext
isiteleka = strike
isifanekiso = template
isithombe = picture
# i-doctype = doctype
ukuphawula = comment
